/* ez-bitset.c: Moderately Scalable set of 32-bit integers.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#ifndef _EZ_BITSET_H
#define _EZ_BITSET_H

typedef struct ez_bitset ez_bitset;

typedef struct ez_bitset_scan ez_bitset_scan;

/**
 * Scan state.
 *
 * This is public so it can be allocated on the stack, but the fields are opaque.
 */
struct ez_bitset_scan {
	ez_bitset *set;
	void *l1, *l2;
	int id0, id1, id2;
	uint32_t offset, bits, bit;
};

/**
 * Allocate a new bitset.
 */
ez_bitset *ez_bitset_new(void);

/**
 * Clear all set bits from the bitset.
 */
void ez_bitset_clear(ez_bitset *bs);

/**
 * Free a bitset.
 */
void ez_bitset_free(ez_bitset *bs);

/**
 * Get a bit by index.
 */
int ez_bitset_get(ez_bitset *bs, uint32_t bit);

/**
 * Set a bit by index to a particular value.
 *
 * @return the previous value for this bit.
 */
int ez_bitset_set(ez_bitset *bs, uint32_t bit, int value);

/**
 * Return the cardinality of the set.  The number of set bits.
 */
uint32_t ez_bitset_card(ez_bitset *bs);

/**
 * Initiate an iterative scan of all bits.  Bits are scanned in order.
 *
 * If the bitset contains ~0 then a scan will ignore it.
 *
 * @return The first bit set.  ~0 indicates none found.
 */
uint32_t ez_bitset_scan_init(ez_bitset *set, ez_bitset_scan *scan);

/**
 * Move to the next bit.
 *
 * @return the next set bit or ~0.
 */
uint32_t ez_bitset_scan_next(ez_bitset_scan *scan);

/**
 * Clear this bit and move to the next one.
 *
 * This must only be called when a set bit was found in a previous
 * scan call.
 *
 * @return the next set bit or ~0.
 */
uint32_t ez_bitset_scan_remove(ez_bitset_scan *scan);

#endif
