
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <assert.h>

#include "ez-blob.h"
#include "ez-list.h"

#include "ez-blob-blob.h"
#include "ez-blob-tagz.h"
#include "ez-blob-xdrn.h"

struct simple {
	char b;
	short s;
	int a;
	float c;
	double d;
};

static const ez_blob_desc simple_DESC[] = {
	EZ_BLOB_START(struct simple, 0, 5),
	EZ_BLOB_INT8(struct simple, 1, b),
	EZ_BLOB_INT16(struct simple, 2, s),
	EZ_BLOB_INT32(struct simple, 3, a),
	EZ_BLOB_FLOAT32(struct simple, 4, c),
	EZ_BLOB_FLOAT64(struct simple, 5, d),
};

struct arrays {
	uint32_t count;
	char *string;
	ez_blob array;
};

static const ez_blob_desc array_DESC[] = {
	EZ_BLOB_START(struct arrays, 0, 2),
	EZ_BLOB_STRING(struct arrays, 1, string),
	EZ_BLOB_BLOB(struct arrays, 2, array),
};

struct embedded {
	struct simple s;
	struct arrays a;
};

static const ez_blob_desc embed_DESC[] = {
	EZ_BLOB_START(struct embedded, 0, 2),
	EZ_BLOB_STRUCT(struct embedded, 1, s, simple_DESC),
	EZ_BLOB_STRUCT(struct embedded, 2, a, array_DESC),
};

struct tree {
	struct tree *left;
	struct tree *right;

	char *s;
};

static const ez_blob_desc tree_DESC[] = {
	EZ_BLOB_START(struct tree, 0, 3),
	EZ_BLOB_STRING(struct tree, 1, s),
	EZ_BLOB_POINTER_NULL(struct tree, 2, left, tree_DESC),
	EZ_BLOB_POINTER_NULL(struct tree, 3, right, tree_DESC),
};

struct listnode {
	ez_node ln;
	char *name;
	ez_blob value;
};

static const ez_blob_desc listnode_DESC[] = {
	EZ_BLOB_START(struct listnode, 0, 2),
	EZ_BLOB_STRING(struct listnode, 1, name),
	EZ_BLOB_BLOB(struct listnode, 2, value),
};

struct list {
	char *name;
	ez_list list;
};

static const ez_blob_desc list_DESC[] = {
	EZ_BLOB_START(struct list, 0, 2),
	EZ_BLOB_STRING(struct list, 1, name),
	EZ_BLOB_LIST(struct list, 2, list, listnode_DESC),
};

static void dumphex(const char *data, size_t size, const char *prefix) {
	for (int i=0;i<size;i+=16) {
		fputs(prefix, stdout);
		printf(" %04x:", i);
		for (int j=0;j<16;j++) {
			if (i+j<size)
				printf(" %02x", data[i+j] & 0xff);
			else
				fputs("   ", stdout);
		}
		fputs("  ", stdout);
		for (int j=0;j<16;j++) {
			putchar(i+j<size && isprint(data[i+j]) ? data[i+j] : '.');
		}
		putchar('\n');
	}
}

static void blob_print(const ez_blob_desc *d, const void *a, int depth) {
	ez_blob_print(d, a, depth);
}

static int blob_equals(const ez_blob_desc *desc, const void *a, const void *b) {
	int e = 1;
	for (int i=0, dlen=desc->bd_length; e && i < dlen; i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *u = a + d->bd_offset;
		const void *v = b + d->bd_offset;

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			e &= (*(uint8_t *)u) == (*(uint8_t *)v);
			break;
		case EZ_BLOB_INT16:
			e &= (*(uint16_t *)u) == (*(uint16_t *)v);
			break;
		case EZ_BLOB_INT32:
			e &= (*(uint32_t *)u) == (*(uint32_t *)v);
			break;
		case EZ_BLOB_INT64:
			e &= (*(uint64_t *)u) == (*(uint64_t *)v);
			break;
		case EZ_BLOB_FLOAT32:
			e &= (*(float *)u) == (*(float *)v);
			break;
		case EZ_BLOB_FLOAT64:
			e &= (*(double *)u) == (*(double *)v);
			break;
		case EZ_BLOB_STRING | EZ_BLOB_NULLABLE:
			if (!*(char **)u || !*(char **)v) {
				e &= (*(char **)u) == (*(char **)v);
				break;
			}
			// falls through
		case EZ_BLOB_STRING:
			if (!*(char **)u || !*(char **)v) {
				printf("strings can't be null\n");
				e = 0;
			} else {
				e &= strcmp(*(char **)u, *(char **)v) == 0;
			}
			break;
		case EZ_BLOB_BLOB: {
			const struct ez_blob *ua = u;
			const struct ez_blob *va = v;

			e &= ua->size == va->size
				&& memcmp(ua->data, va->data, ua->size) == 0;
			break;
		}
		case EZ_BLOB_POINTER | EZ_BLOB_NULLABLE:
			if (!*(void **)u || !*(void **)v) {
				e &= *(void **)u == *(void **)v;
				break;
			}
			// falls through
		case EZ_BLOB_POINTER:
			e &= blob_equals(d->bd_table, *(void **)u, *(void **)v);
			break;
		case EZ_BLOB_LIST: {
			ez_list *ul = (void *)u;
			ez_list *vl = (void *)v;

			e &= ez_list_size(ul) == ez_list_size(vl);
			for (ez_node *uw = ez_list_head(ul), *un = ez_node_succ(uw),
				     *vw = ez_list_head(vl), *vn = ez_node_succ(uw);
			     un && vn;
			     uw = un, un = ez_node_succ(un),
				     vw = vn, vn = ez_node_succ(vn)) {
				e &= blob_equals(d->bd_table, uw, vw);
			}
			
			break;
		}
		default:
			printf("cmp unknown field type=%d id=%d\n", d->bd_type, d->bd_id);
			e = 0;
			break;
		}
		if (!e) {
			printf("cmp failed in struct %d at field %d type %04x\n", desc->bd_id, d->bd_id, d->bd_type);
			abort();
		}
	}

	return e;
}

struct test_funcs {
	const char *name;
	int (*blob_encode)(const ez_blob_desc *d, const void *p, ez_blob *blob);
	void *(*blob_decode)(const ez_blob_desc *desc, const ez_blob *blob);
	int (*blob_decode_raw)(const ez_blob_desc *desc, const ez_blob *blob, void *p);
};

static void test_basics(const ez_blob_desc *d, void *src, struct test_funcs *funcs) {
	ez_blob blob, blob2;
	void *t;
	size_t dsize = d->bd_offset;
	char tmp[dsize + 64] __attribute__ ((aligned(64)));
	int res;
	
	printf("source\n");
	blob_print(d, src, 4);
	
	res = funcs->blob_encode(d, src, &blob);
	assert(res == 0);
	dumphex(blob.data, blob.size, "serial: ");

	// check decode equals
	t = funcs->blob_decode(d, &blob);
	assert(t != NULL);
	assert(blob_equals(d, t, src));

	printf("decoded\n");
	blob_print(d, t, 4);

	// check encode-decoded equals source
	funcs->blob_encode(d, t, &blob2);
	assert(blob.size == blob2.size);
	assert(memcmp(blob.data, blob2.data, blob2.size) == 0);
	free(blob2.data);
	
	ez_blob_free(d, t);

	// check raw decode stays within bounds of struct
	memset(tmp, 0xbe, sizeof(tmp));
	funcs->blob_decode_raw(d, &blob, &tmp[32]);
	//assert(t != NULL);
	//assert(blob_equals(d, tmp+32, src));
	dumphex(tmp, sizeof(tmp), "object: ");

	for (int i=0;i<32;i++) {
		assert((tmp[i] & 0xff) == 0xbe);
		assert((tmp[i + 32 + dsize] & 0xff) == 0xbe);
	}
	ez_blob_free_raw(d, tmp+32);
	free(blob.data);
}

static void test_simple(struct test_funcs *funcs) {
	struct simple src = {
		'z', 0xff0f, 1, 3.14159f, 5644941221133
	};
	const ez_blob_desc *d = simple_DESC;

	printf("test simple\n");
	test_basics(d, &src, funcs);
}

static void test_arrays(struct test_funcs *funcs) {
	char data[7] = { 1, 2, 3, 4, 5, 6, 7 };
	struct arrays src = {
		32,
		"Bob was here.",
		.array.size = 7,
		.array.data = data
	};
	const ez_blob_desc *d = array_DESC;

	printf("test arrays\n");
	test_basics(d, &src, funcs);
}

static void test_rawtree(struct test_funcs *funcs) {
	struct tree src[6] = {
		{ &src[1], &src[3], "root" },
		{ NULL, &src[2], "left" },
		{ NULL, NULL, "left-right" },
		{ &src[4], &src[5], "right" },
		{ NULL, NULL, "right-left" },
		{ NULL, NULL, "right-right" },
	};
	const ez_blob_desc *d = tree_DESC;

	printf("test raw tree\n");
	test_basics(d, &src, funcs);
}

static void test_list(struct test_funcs *funcs) {
	struct list list = {
		.name = "a list",
		.list = EZ_INIT_LIST(list.list)
	};
	struct listnode nodes[3] = {
		{ .name = "node 0", .value.size = 5, .value.data = "data0" },
		{ .name = "node 1", .value.size = 5, .value.data = "data1" },
		{ .name = "node 2", .value.size = 5, .value.data = "data2" },
	};

	list.name = "a list";
	//	ez_list_init(&list.
	for (int i=0;i<3;i++)
		ez_list_addtail(&list.list, &nodes[i]);
	
	const ez_blob_desc *d = list_DESC;

	printf("test list\n");
	test_basics(d, &list, funcs);
}

struct test_funcs backends[] = {
	{
		"blob legacy",
		ez_blob_encode,
		ez_blob_decode,
		ez_blob_decode_raw
	},
	{
		"tagz",
		ez_tagz_encode,
		ez_tagz_decode,
		ez_tagz_decode_raw
	},
	{
		"xdrn",
		ez_xdrn_encode,
		ez_xdrn_decode,
		ez_xdrn_decode_raw
	},	
};

int main(int argc, char **argv) {
	for (int i=0;i<3;i++) {
		struct test_funcs *f = &backends[i];
		printf("* **********************************************************************\n");
		printf("test: %s\n", f->name);
		fflush(stdout);

		test_simple(f);
		test_arrays(f);
		test_rawtree(f);
		test_list(f);
	}
}
